CHIMERA_MODEL_WORKER_FIT_PATH = "/v1/chimera-model/fit"
CHIMERA_MODEL_WORKER_PREDICT_PATH = "/v1/chimera-model/predict"

CHIMERA_SGD_WORKER_FIT_STEP_PATH = "/v1/chimera-sgd/fit-step"
CHIMERA_SGD_WORKER_FIT_REQUEST_DATA_SAMPLE_PATH = (
    "/v1/chimera-sgd/request-data-sample"
)

CHIMERA_AGGREGATION_MASTER_FIT_PATH = "/v1/chimera-aggregation/fit"
CHIMERA_AGGREGATION_MASTER_PREDICT_PATH = "/v1/chimera-aggregation/predict"

CHIMERA_PARAMETER_SERVER_MASTER_FIT_PATH = "/v1/chimera-parameter-server/fit"
CHIMERA_PARAMETER_SERVER_MASTER_PREDICT_PATH = "/v1/chimera-parameter-server/predict"
